@@ -590,9 +590,7 @@ func Load(arch *sys.Arch, syms *sym.Symbols, f *bio.Reader, pkg string, length i
 
 	elfobj.nsect = uint(elfobj.shnum)
 	for i := 0; uint(i) < elfobj.nsect; i++ {
-		if f.Seek(int64(uint64(base)+elfobj.shoff+uint64(int64(i)*int64(elfobj.shentsize))), 0) < 0 {
-			return errorf("malformed elf file: negative seek")
-		}
+		f.MustSeek(int64(uint64(base)+elfobj.shoff+uint64(int64(i)*int64(elfobj.shentsize))), 0)
 		sect := &elfobj.sect[i]
 		if is64 != 0 {
 			var b ElfSectBytes64
@@ -794,6 +792,13 @@ func Load(arch *sys.Arch, syms *sym.Symbols, f *bio.Reader, pkg string, length i
 				continue
 			}
 
+			if strings.HasPrefix(elfsym.name, "$d") && elfsym.type_ == 0 && sect.name == ".debug_frame" {
+				// "$d" is a marker, not a real symbol.
+				// This happens with gcc on ARM64.
+				// See https://sourceware.org/bugzilla/show_bug.cgi?id=21809
+				continue
+			}
+
 			if strings.HasPrefix(elfsym.name, ".LASF") { // gcc on s390x does this
 				continue
 			}
@@ -931,7 +936,7 @@ func Load(arch *sys.Arch, syms *sym.Symbols, f *bio.Reader, pkg string, length i
 				rp.Sym = elfsym.sym
 			}
 
-			rp.Type = 256 + objabi.RelocType(info)
+			rp.Type = objabi.ElfRelocOffset + objabi.RelocType(info)
 			rp.Siz, err = relSize(arch, pn, uint32(info))
 			if err != nil {
 				return nil, 0, err
@@ -989,9 +994,7 @@ func elfmap(elfobj *ElfObj, sect *ElfSect) (err error) {
 	}
 
 	sect.base = make([]byte, sect.size)
-	if elfobj.f.Seek(int64(uint64(elfobj.base)+sect.off), 0) < 0 {
-		return fmt.Errorf("short read: seek not successful")
-	}
+	elfobj.f.MustSeek(int64(uint64(elfobj.base)+sect.off), 0)
 	if _, err := io.ReadFull(elfobj.f, sect.base); err != nil {
 		return fmt.Errorf("short read: %v", err)
 	}
@@ -1064,8 +1067,8 @@ func readelfsym(arch *sys.Arch, syms *sym.Symbols, elfobj *ElfObj, i int, elfsym
 			}
 
 		case ElfSymBindLocal:
-			if arch.Family == sys.ARM && (strings.HasPrefix(elfsym.name, "$a") || strings.HasPrefix(elfsym.name, "$d")) {
-				// binutils for arm generate these mapping
+			if (arch.Family == sys.ARM || arch.Family == sys.ARM64) && (strings.HasPrefix(elfsym.name, "$a") || strings.HasPrefix(elfsym.name, "$d") || strings.HasPrefix(elfsym.name, "$x")) {
+				// binutils for arm and arm64 generate these mapping
 				// symbols, ignore these
 				break
 			}
@@ -1096,6 +1099,11 @@ func readelfsym(arch *sys.Arch, syms *sym.Symbols, elfobj *ElfObj, i int, elfsym
 				if elfsym.other == 2 {
 					s.Attr |= sym.AttrVisibilityHidden
 				}
+
+				// Allow weak symbols to be duplicated when already defined.
+				if s.Outer != nil {
+					s.Attr |= sym.AttrDuplicateOK
+				}
 			}
 
 		default:
@@ -1122,81 +1130,95 @@ func relSize(arch *sys.Arch, pn string, elftype uint32) (uint8, error) {
 	const (
 		AMD64 = uint32(sys.AMD64)
 		ARM   = uint32(sys.ARM)
+		ARM64 = uint32(sys.ARM64)
 		I386  = uint32(sys.I386)
 		PPC64 = uint32(sys.PPC64)
 		S390X = uint32(sys.S390X)
 	)
 
-	switch uint32(arch.Family) | elftype<<24 {
+	switch uint32(arch.Family) | elftype<<16 {
 	default:
 		return 0, fmt.Errorf("%s: unknown relocation type %d; compiled without -fpic?", pn, elftype)
 
-	case S390X | uint32(elf.R_390_8)<<24:
+	case S390X | uint32(elf.R_390_8)<<16:
 		return 1, nil
 
-	case PPC64 | uint32(elf.R_PPC64_TOC16)<<24,
-		PPC64 | uint32(elf.R_PPC64_TOC16_LO)<<24,
-		PPC64 | uint32(elf.R_PPC64_TOC16_HI)<<24,
-		PPC64 | uint32(elf.R_PPC64_TOC16_HA)<<24,
-		PPC64 | uint32(elf.R_PPC64_TOC16_DS)<<24,
-		PPC64 | uint32(elf.R_PPC64_TOC16_LO_DS)<<24,
-		PPC64 | uint32(elf.R_PPC64_REL16_LO)<<24,
-		PPC64 | uint32(elf.R_PPC64_REL16_HI)<<24,
-		PPC64 | uint32(elf.R_PPC64_REL16_HA)<<24,
-		S390X | uint32(elf.R_390_16)<<24,
-		S390X | uint32(elf.R_390_GOT16)<<24,
-		S390X | uint32(elf.R_390_PC16)<<24,
-		S390X | uint32(elf.R_390_PC16DBL)<<24,
-		S390X | uint32(elf.R_390_PLT16DBL)<<24:
+	case PPC64 | uint32(elf.R_PPC64_TOC16)<<16,
+		PPC64 | uint32(elf.R_PPC64_TOC16_LO)<<16,
+		PPC64 | uint32(elf.R_PPC64_TOC16_HI)<<16,
+		PPC64 | uint32(elf.R_PPC64_TOC16_HA)<<16,
+		PPC64 | uint32(elf.R_PPC64_TOC16_DS)<<16,
+		PPC64 | uint32(elf.R_PPC64_TOC16_LO_DS)<<16,
+		PPC64 | uint32(elf.R_PPC64_REL16_LO)<<16,
+		PPC64 | uint32(elf.R_PPC64_REL16_HI)<<16,
+		PPC64 | uint32(elf.R_PPC64_REL16_HA)<<16,
+		S390X | uint32(elf.R_390_16)<<16,
+		S390X | uint32(elf.R_390_GOT16)<<16,
+		S390X | uint32(elf.R_390_PC16)<<16,
+		S390X | uint32(elf.R_390_PC16DBL)<<16,
+		S390X | uint32(elf.R_390_PLT16DBL)<<16:
 		return 2, nil
 
-	case ARM | uint32(elf.R_ARM_ABS32)<<24,
-		ARM | uint32(elf.R_ARM_GOT32)<<24,
-		ARM | uint32(elf.R_ARM_PLT32)<<24,
-		ARM | uint32(elf.R_ARM_GOTOFF)<<24,
-		ARM | uint32(elf.R_ARM_GOTPC)<<24,
-		ARM | uint32(elf.R_ARM_THM_PC22)<<24,
-		ARM | uint32(elf.R_ARM_REL32)<<24,
-		ARM | uint32(elf.R_ARM_CALL)<<24,
-		ARM | uint32(elf.R_ARM_V4BX)<<24,
-		ARM | uint32(elf.R_ARM_GOT_PREL)<<24,
-		ARM | uint32(elf.R_ARM_PC24)<<24,
-		ARM | uint32(elf.R_ARM_JUMP24)<<24,
-		AMD64 | uint32(elf.R_X86_64_PC32)<<24,
-		AMD64 | uint32(elf.R_X86_64_PLT32)<<24,
-		AMD64 | uint32(elf.R_X86_64_GOTPCREL)<<24,
-		AMD64 | uint32(elf.R_X86_64_GOTPCRELX)<<24,
-		AMD64 | uint32(elf.R_X86_64_REX_GOTPCRELX)<<24,
-		I386 | uint32(elf.R_386_32)<<24,
-		I386 | uint32(elf.R_386_PC32)<<24,
-		I386 | uint32(elf.R_386_GOT32)<<24,
-		I386 | uint32(elf.R_386_PLT32)<<24,
-		I386 | uint32(elf.R_386_GOTOFF)<<24,
-		I386 | uint32(elf.R_386_GOTPC)<<24,
-		I386 | uint32(elf.R_386_GOT32X)<<24,
-		PPC64 | uint32(elf.R_PPC64_REL24)<<24,
-		PPC64 | uint32(elf.R_PPC_REL32)<<24,
-		S390X | uint32(elf.R_390_32)<<24,
-		S390X | uint32(elf.R_390_PC32)<<24,
-		S390X | uint32(elf.R_390_GOT32)<<24,
-		S390X | uint32(elf.R_390_PLT32)<<24,
-		S390X | uint32(elf.R_390_PC32DBL)<<24,
-		S390X | uint32(elf.R_390_PLT32DBL)<<24,
-		S390X | uint32(elf.R_390_GOTPCDBL)<<24,
-		S390X | uint32(elf.R_390_GOTENT)<<24:
+	case ARM | uint32(elf.R_ARM_ABS32)<<16,
+		ARM | uint32(elf.R_ARM_GOT32)<<16,
+		ARM | uint32(elf.R_ARM_PLT32)<<16,
+		ARM | uint32(elf.R_ARM_GOTOFF)<<16,
+		ARM | uint32(elf.R_ARM_GOTPC)<<16,
+		ARM | uint32(elf.R_ARM_THM_PC22)<<16,
+		ARM | uint32(elf.R_ARM_REL32)<<16,
+		ARM | uint32(elf.R_ARM_CALL)<<16,
+		ARM | uint32(elf.R_ARM_V4BX)<<16,
+		ARM | uint32(elf.R_ARM_GOT_PREL)<<16,
+		ARM | uint32(elf.R_ARM_PC24)<<16,
+		ARM | uint32(elf.R_ARM_JUMP24)<<16,
+		ARM64 | uint32(elf.R_AARCH64_CALL26)<<16,
+		ARM64 | uint32(elf.R_AARCH64_ADR_GOT_PAGE)<<16,
+		ARM64 | uint32(elf.R_AARCH64_LD64_GOT_LO12_NC)<<16,
+		ARM64 | uint32(elf.R_AARCH64_ADR_PREL_PG_HI21)<<16,
+		ARM64 | uint32(elf.R_AARCH64_ADD_ABS_LO12_NC)<<16,
+		ARM64 | uint32(elf.R_AARCH64_LDST8_ABS_LO12_NC)<<16,
+		ARM64 | uint32(elf.R_AARCH64_LDST32_ABS_LO12_NC)<<16,
+		ARM64 | uint32(elf.R_AARCH64_LDST64_ABS_LO12_NC)<<16,
+		ARM64 | uint32(elf.R_AARCH64_LDST128_ABS_LO12_NC)<<16,
+		ARM64 | uint32(elf.R_AARCH64_PREL32)<<16,
+		ARM64 | uint32(elf.R_AARCH64_JUMP26)<<16,
+		AMD64 | uint32(elf.R_X86_64_PC32)<<16,
+		AMD64 | uint32(elf.R_X86_64_PLT32)<<16,
+		AMD64 | uint32(elf.R_X86_64_GOTPCREL)<<16,
+		AMD64 | uint32(elf.R_X86_64_GOTPCRELX)<<16,
+		AMD64 | uint32(elf.R_X86_64_REX_GOTPCRELX)<<16,
+		I386 | uint32(elf.R_386_32)<<16,
+		I386 | uint32(elf.R_386_PC32)<<16,
+		I386 | uint32(elf.R_386_GOT32)<<16,
+		I386 | uint32(elf.R_386_PLT32)<<16,
+		I386 | uint32(elf.R_386_GOTOFF)<<16,
+		I386 | uint32(elf.R_386_GOTPC)<<16,
+		I386 | uint32(elf.R_386_GOT32X)<<16,
+		PPC64 | uint32(elf.R_PPC64_REL24)<<16,
+		PPC64 | uint32(elf.R_PPC_REL32)<<16,
+		S390X | uint32(elf.R_390_32)<<16,
+		S390X | uint32(elf.R_390_PC32)<<16,
+		S390X | uint32(elf.R_390_GOT32)<<16,
+		S390X | uint32(elf.R_390_PLT32)<<16,
+		S390X | uint32(elf.R_390_PC32DBL)<<16,
+		S390X | uint32(elf.R_390_PLT32DBL)<<16,
+		S390X | uint32(elf.R_390_GOTPCDBL)<<16,
+		S390X | uint32(elf.R_390_GOTENT)<<16:
 		return 4, nil
 
-	case AMD64 | uint32(elf.R_X86_64_64)<<24,
-		AMD64 | uint32(elf.R_X86_64_PC64)<<24,
-		PPC64 | uint32(elf.R_PPC64_ADDR64)<<24,
-		S390X | uint32(elf.R_390_GLOB_DAT)<<24,
-		S390X | uint32(elf.R_390_RELATIVE)<<24,
-		S390X | uint32(elf.R_390_GOTOFF)<<24,
-		S390X | uint32(elf.R_390_GOTPC)<<24,
-		S390X | uint32(elf.R_390_64)<<24,
-		S390X | uint32(elf.R_390_PC64)<<24,
-		S390X | uint32(elf.R_390_GOT64)<<24,
-		S390X | uint32(elf.R_390_PLT64)<<24:
+	case AMD64 | uint32(elf.R_X86_64_64)<<16,
+		AMD64 | uint32(elf.R_X86_64_PC64)<<16,
+		ARM64 | uint32(elf.R_AARCH64_ABS64)<<16,
+		ARM64 | uint32(elf.R_AARCH64_PREL64)<<16,
+		PPC64 | uint32(elf.R_PPC64_ADDR64)<<16,
+		S390X | uint32(elf.R_390_GLOB_DAT)<<16,
+		S390X | uint32(elf.R_390_RELATIVE)<<16,
+		S390X | uint32(elf.R_390_GOTOFF)<<16,
+		S390X | uint32(elf.R_390_GOTPC)<<16,
+		S390X | uint32(elf.R_390_64)<<16,
+		S390X | uint32(elf.R_390_PC64)<<16,
+		S390X | uint32(elf.R_390_GOT64)<<16,
+		S390X | uint32(elf.R_390_PLT64)<<16:
 		return 8, nil
 	}
 }
